program s
  use input_mdl, only: muUV,muIR,muMS,dp
  use modelinput, only: linc,printp,rs,scalefac
  use pde_integrator, only: rk4_step, euler_step, update_func, set_kernel
  use class_particles, only: print_parameters_mdl
  implicit none
  character(len=10) :: renoscheme = ''
  complex(dp) :: t,Gf,minv,minmh2,minmsb2
  integer :: i
  real(dp) :: cab_in,tb_in,mhh_in,mha_in,mhc_in,l5_in
  integer :: ytype_in

  integer, parameter :: steps = 400
  real(dp), parameter :: scale_init = 100d0
  real(dp), dimension(-steps-1:steps+1) :: scale_arr,cab_arr,tb_arr,l5_arr

  procedure (update_func), pointer :: reset_func
  procedure (set_kernel), pointer  :: set_k

  character(len=6) :: benchmarkpoint = 'run'
  character(len=100) :: file_bp
  character(len=*), parameter :: fmt = '(e21.14,x,e21.14,x,e21.14,x,e21.14)'
  complex(dp),parameter :: cone=1d0,cnul=0d0

  reset_func => reset_alpha_beta
  set_k => set_beta_alpha_beta

  call read_input

  renoscheme = trim(rs)

  if (printp .gt. 0) call print_parameters_mdl(printp-1)

  call reset_alpha_beta
  call running(scale_init,scale_init/scalefac)
  call reset_thdm_param
  call reset_alpha_beta
  call running(scale_init,scale_init*scalefac)

  file_bp = trim(benchmarkpoint) // '_' // trim(renoscheme)
  open(unit=1,file=trim(file_bp),status='replace',position='append')

  write(1,'(a,14x,a,16x,a,14x,a)') '# mu/mu0', 'tb(mu)', 'cab(mu)', 'l5(mu)'
  do i = -steps, steps
    if (i .ne. 0) then
      write(1,fmt) scale_arr(i), tb_arr(i), cab_arr(i), l5_arr(i)
    else
      write(1,fmt) 1d0, tb_in, cab_in, l5_in
    end if
  end do

  close(1)
  close(2)

  contains

  subroutine read_input
    use fill_couplings, only: fill_couplings_mdl
    use thdm_ct, only: initct
    use collier_interface, only: init_collier
    use modelinput, only: read_modelinput,cab,tb,l5,MHH,MHA,MHC,printp,ytype,scalefac
    use modelfile, only: set_parameter_mdl
    character(len=*), parameter :: fmt0 = '(a,f21.15)'
    character(len=*), parameter :: fmt = '(2x,a,x,f23.14)'
    character(len=*), parameter :: fmt_i = '(2x,a,x,i1)'
    character(len=*), parameter :: fmt_a = '(2x,a,x,a)'

    call read_modelinput

    muUV = 100d0
    muIR = 100d0
    muMS = 100d0

    mhh_in = MHH
    mha_in = MHA
    mhc_in = MHC
    cab_in = cab
    tb_in = tb
    l5_in = l5
    ytype_in = ytype

    if (printp .eq. 0) then
      write(*,fmt) "mhh in:", mhh_in
      write(*,fmt) "mha in:", mha_in
      write(*,fmt) "mhc in:", mhc_in
      write(*,fmt) "cab in:", cab_in
      write(*,fmt) "tb  in:", tb_in
      write(*,fmt) "l5  in:", l5_in
      write(*,fmt_i) "ytype in:", ytype_in
      write(*,fmt_a) "rs:", trim(rs)
    end if

    tb_arr(0) = tb_in
    cab_arr(0) = cab_in
    l5_arr(0) = l5_in
    scale_arr(0) = 1d0

    call reset_thdm_param

    muMS = (125d0 + mhh_in + mha_in + 2*mhc_in)/5d0
    muUV = muMS
    call set_parameter_mdl('muMS_BSM', muMS*cone)

    write(*,fmt0) " tb(mu0)", tb_in
    write(*,fmt0) "cab(mu0)", cab_in
    write(*,fmt0) " l5(mu0)", l5_in

    call init_collier

  end subroutine read_input

  subroutine reset_thdm_param()
    use modelfile, only: set_parameter_mdl
    call set_tb_cab_rcl(tb_in,cab_in)
    call set_Z2_thdm_yukawa_type_rcl(ytype_in)
    call set_parameter_mdl('l5', l5_in*cone)
    call set_parameter_mdl('MHH', mhh_in*cone)
    call set_parameter_mdl("WHH",cnul)
    call set_parameter_mdl('MHC', mhc_in*cone)
    call set_parameter_mdl("WHC",cnul)
    call set_parameter_mdl('MHA', mha_in*cone)
    call set_parameter_mdl("WHA",cnul)
  end subroutine reset_thdm_param

  subroutine running(scale0,scale)
    use class_particles, only: sa,cb,l5,cab,tb,pi
    use pde_integrator, only: rk4_step, euler_step
    real(dp), intent(in)    :: scale0,scale
    real(dp) :: a0,b0,l50,ee0,am,bm,l5m,eem,t0,t,tnlo
    complex(dp),dimension(3) :: y
    real(dp) :: scale20,scale2,stepsize,scale2m,tbr,cabr
    integer :: step
    character(len=30) :: fmt = '(a,f4.2,a,f21.15)'

    scale20 = scale0**2d0
    scale2  = scale**2d0
    step = 0

    stepsize = (scale2-scale20)/real(steps,kind=dp)

    tbr=tb
    cabr=cab
    b0 = atan(tbr)
    a0 = acos(cabr) + b0
    l50 = l5

    if (a0 .gt. pi/2d0) then
      a0 = 2*b0 - a0
    end if

    call reset_alpha_beta

    scale2m = scale20
    am      = a0
    bm      = b0
    l5m     = l50
    do while ((scale2m .gt. scale2 .and. stepsize .lt. 0d0) .or. &
              (scale2m .lt. scale2 .and. stepsize .gt. 0d0))
      y = (/am, bm, l5m/)
      call rk4_step(y,scale2m,stepsize,3,reset_func,set_k)
      !call euler_step(y,scale2m,stepsize)
      am = y(1)
      bm = y(2)
      l5m = y(3)

      if (am .gt. pi/2d0) then
        am = 2*bm - am
      end if

      sa = sin(am)
      cb = cos(bm)
      l5 = l5m

      call reset_alpha_beta

      if (stepsize .gt. 0) then
        step = step + 1
      else
        step = step - 1
      end if
      tb_arr(step) = tan(bm)
      cab_arr(step) = cos(am-bm)
      l5_arr(step) = l5m
      scale_arr(step) = sqrt(scale2m)/scale_init
    end do
    sa = sin(am)
    cb = cos(bm)
    l5 = l5m*cone

    call reset_alpha_beta

    if (scale .lt. scale0) then
      write(*,fmt) " tb(mu0/", real(scale0/scale), ")", tan(bm)
      write(*,fmt) "cab(mu0/", real(scale0/scale), ")", cos(am-bm)
      write(*,fmt) " l5(mu0/", real(scale0/scale), ")", l5m
    else
      write(*,fmt) " tb(", real(scale/scale0), " mu0)", tan(bm)
      write(*,fmt) "cab(", real(scale/scale0), " mu0)", cos(am-bm)
      write(*,fmt) " l5(", real(scale/scale0), " mu0)", l5m
    end if

  end subroutine running

  subroutine set_tb_cab_rcl(tb,cab)
    use class_particles, only: pi,cb,sa
    real(dp), intent(in) :: tb,cab
    real(dp)             :: b,a

    b = atan(tb)
    cb = cos(b)
    a = acos(cab) + b

    if (a .gt. pi/2d0) then
      a = 2*b - a
    end if
    sa = sin(a)
  end subroutine set_tb_cab_rcl

  subroutine set_Z2_thdm_yukawa_type_rcl (ytype)
    use modelfile, only: set_parameter_mdl

    ! This subroutine sets the yukawa type in the softly-broken Z2 symmetric
    ! THDM to "ytype".
    ! ytype = 1 : Type-I   h1u=h1d=h1l=0
    ! ytype = 2 : Type-II  h1u=h2d=h2l=0
    ! ytype = 3 : Type-X   h1u=h1d=h2l=0
    ! ytype = 4 : Type-Y   h1u=h2d=h1l=0

    integer, intent(in) :: ytype
    select case (ytype)
    case (1)
      call set_parameter_mdl("h1u", cnul)
      call set_parameter_mdl("h2u", cone)

      call set_parameter_mdl("h1d", cnul)
      call set_parameter_mdl("h2d", cone)

      call set_parameter_mdl("h1l", cnul)
      call set_parameter_mdl("h2l", cone)
    case (2)
      call set_parameter_mdl("h1u", cnul)
      call set_parameter_mdl("h2u", cone)

      call set_parameter_mdl("h2d", cnul)
      call set_parameter_mdl("h1d", cone)

      call set_parameter_mdl("h2l", cnul)
      call set_parameter_mdl("h1l", cone)
    case (3)
      call set_parameter_mdl("h1u", cnul)
      call set_parameter_mdl("h2u", cone)

      call set_parameter_mdl("h1d", cnul)
      call set_parameter_mdl("h2d", cone)

      call set_parameter_mdl("h2l", cnul)
      call set_parameter_mdl("h1l", cone)
    case (4)
      call set_parameter_mdl("h1u", cnul)
      call set_parameter_mdl("h2u", cone)

      call set_parameter_mdl("h2d", cnul)
      call set_parameter_mdl("h1d", cone)

      call set_parameter_mdl("h1l", cnul)
      call set_parameter_mdl("h2l", cone)
    end select

  end subroutine set_Z2_thdm_yukawa_type_rcl

  subroutine reset_alpha_beta(yc)
    use class_particles, only: sa,cb,l5,msb2,init_parameters_mdl
    use thdm_ct, only: initct, da_FJTS, db_FJTS, da_PRTS, db_PRTS, dl5_FJTS
    complex(dp), dimension(:), optional, intent(in) :: yc
    complex(dp) :: da_PRTS_tmp, db_PRTS_tmp, dl5_FJTS_tmp

    if (present(yc)) then
      sa = sin(yc(1))
      cb = cos(yc(2))
      l5 = real(yc(3),kind=dp)
    endif

    call init_parameters_mdl()
    call initct(0d0)
    dl5_FJTS_tmp = dl5_FJTS
    da_PRTS_tmp = da_PRTS
    db_PRTS_tmp = db_PRTS
    call initct(1d0)
    dl5_FJTS = dl5_FJTS - dl5_FJTS_tmp
    da_PRTS = da_PRTS - da_PRTS_tmp
    db_PRTS = db_PRTS - db_PRTS_tmp

  end subroutine reset_alpha_beta

  subroutine set_beta_alpha_beta(x,f)

    use class_particles
    use thdm_ct, only: da_FJTS, db_FJTS, da_PRTS, db_PRTS, dl5_FJTS
    real(dp),                  intent(in)  :: x
    complex(dp), dimension(:), intent(out) :: f

    if (renoscheme .eq. 'FJ') then
      f(:) = (/da_FJTS/x,db_FJTS/x,dl5_FJTS/x/)
    else if (renoscheme .eq. 'PR') then
      f(:) = (/da_PRTS/x,db_PRTS/x,dl5_FJTS/x/)
    else
      f(:) = (/cnul,cnul,dl5_FJTS/x/)
    end if

  end subroutine set_beta_alpha_beta

end program s
