!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                            lbfgsb_interface.f90                             !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module lbfgsb_interface
  implicit none
  integer, parameter  :: dp = kind(1.0d0)
  character(len=60)   :: task, csave
  integer, parameter  :: iprint = -1
  real(dp), parameter :: factr = 1.0d+0, pgtol = 1.0d-10
  logical             :: lsave(4)
  integer             :: isave(44)
  real(dp)            :: dsave(29)

  integer               :: n,m
  integer,  allocatable :: nbd(:),iwa(:)
  real(dp)              :: f
  real(dp), allocatable :: x(:),l(:),u(:),g(:),wa(:)

  procedure(chi2func_ptr),  pointer :: chi2func
  procedure(dchi2func_ptr), pointer :: dchi2func

  abstract interface

    ! pointer to function computing the \chi^2
    subroutine chi2func_ptr(x,res)
      real(kind=kind(23d0)), dimension(:), intent(in)  :: x
      real(kind=kind(23d0)),               intent(out) :: res
    end subroutine chi2func_ptr

    ! pointer to function computing the gradient of chi2func
    subroutine dchi2func_ptr(dim,x,res)
      integer,                               intent(in)  :: dim
      real(kind=kind(23d0)), dimension(dim), intent(in)  :: x
      real(kind=kind(23d0)), dimension(dim), intent(out) :: res
    end subroutine dchi2func_ptr

  end interface

  contains

  subroutine init_grid(nvars,gridsize)
    integer, intent(in) :: nvars,gridsize

    n = nvars
    m = gridsize

    allocate(nbd(n),x(n),l(n),u(n),g(n))
    allocate(iwa(3*n))
    allocate(wa(2*m*n+5*n+11*m*m+8*m))
    
  end subroutine init_grid

  subroutine set_bounds(lower,upper,initial,chi2f,dchi2f)
    real(dp), dimension(:),            intent(in) :: lower, upper, initial
    procedure(chi2func_ptr), pointer,  intent(in) :: chi2f
    procedure(dchi2func_ptr), pointer, intent(in) :: dchi2f

    if (size(lower) .ne. size(l)) then
      write(*,*) "lower's size not equal to l"
      stop
    end if
    if (size(upper) .ne. size(u)) then
      write(*,*) "upper's size not equal to u"
      stop
    end if
    if (size(initial) .ne. size(x)) then
      write(*,*) "initial's size not equal to x"
      stop
    end if
    l = lower
    u = upper
    x = initial
    nbd = 2

    chi2func => chi2f
    dchi2func => dchi2f

  end subroutine set_bounds

  subroutine run_loop()

    task = 'START'
    call dchi2func(n,x,g)
    call chi2func(x,f)

    do while(task(1:2) .eq. 'FG' .or. task .eq. 'NEW_X' .or. task .eq. 'START') 
       
      call setulb(n,m,x,l,u,nbd,f,g,factr,pgtol,wa,iwa,task, &
                  iprint,csave,lsave,isave,dsave)
      if (task(1:2) .eq. 'FG') then
        call chi2func(x,f)
        call dchi2func(n,x,g)
      end if
    end do
    
  end subroutine run_loop

end module lbfgsb_interface
